;===============================================================================
;
;fair hit 2.0
;
;changes the hit/miss system in combat
;
;===============================================================================
header
hirom

;===============================================================================
;
; - clean status when hit (link)
; - chance routines
;
;===============================================================================
org     $C20C2D

;-------------------------------------------------------------------------------
clean_link:
;-------------------------------------------------------------------------------

    ;redirection
        JSL hit_dispel          ;dispel status when hit
        JMP $0C5D               ;continue normal code

;-------------------------------------------------------------------------------
chance_hig:
;-------------------------------------------------------------------------------

    ;get chance
        JSR $4B5A               ;random: [0, 255]
        CMP #$80                ;chance

    ;exit
        RTS

;-------------------------------------------------------------------------------
chance_mid:
;-------------------------------------------------------------------------------

    ;get chance
        JSR $4B5A               ;random: [0, 255]
        CMP #$40                ;chance

    ;exit
        RTS

;-------------------------------------------------------------------------------
chance_low:
;-------------------------------------------------------------------------------

    ;get chance
        JSR $4B5A               ;random: [0, 255]
        CMP #$20                ;chance

    ;exit
        RTS

warnpc  $C20C5D

;===============================================================================
;
; - store evasion instead of hit rate
;
;===============================================================================
org     $C22861

;-------------------------------------------------------------------------------
evasion_store:
;-------------------------------------------------------------------------------

    ;exit
        RTS

warnpc  $C2286D

;===============================================================================
;
; - setup golem hp/hits
;
;===============================================================================
org     $C23F67

;-------------------------------------------------------------------------------
golem_hits:
;-------------------------------------------------------------------------------

    ;setup number of hits
        LDA #$0008              ;note: high byte MUST be zero

warnpc  $C23F6A

;===============================================================================
;
; - main hit or miss system
;
;===============================================================================
org     $C2220D

;-------------------------------------------------------------------------------
hit_main:
;
; main routine
;
;2AB in/out:    fighter identifier (bit)
;1X in/out:     attacker identifier/offset
;1Y in/out:     target identifier/offset
;carry out:     BCC if hit, BCS if miss
;mode in/out    C mode (16-bit), 1XY (8-bit)
;-------------------------------------------------------------------------------

    ;init
        PHX                     ;backup X
        PHA                     ;backup AB
        STZ $EE                 ;clean $EE (flags) and $EF (animation)
        STZ $FE                 ;clean $FE (hit rate) and $FF (evade rate)
        SEP #$20                ;mode 1AB

    ;main routine
        JSR hit_init            ;setup hit and evasion rate
        JSR hit_adjust          ;adjust hit and evsion rate based on conditions
        JSR hit_define          ;define if hit or miss
        JSR miss_animation      ;check and setup miss animation
        JSR miss_dispel         ;check and dispel status when miss
        JSR hit_carry           ;setup carry for callers

    ;exit
        REP #$20                ;mode 2AB
        PLA                     ;restore AB
        PLX                     ;restore X
        RTS                     ;exit

;-------------------------------------------------------------------------------
hit_init:
;
; define hit and evasion rate
;-------------------------------------------------------------------------------

    ;get flags
        JSL flag_main           ;setup main flags

    ;setup hit
        LDA $11A8               ;hit rate
        BEQ .skip7119           ;skip if null
        JSR $4B65               ;random: [0, A-1]
        INC
        STA $FE                 ;save hit rate
.skip7119

    ;setup evasion
        TYX
        LDA $EE                 ;flags
        AND #$01                ;magic/physic flag
        BNE .skip6512           ;X = Y + 0 if physic
        INX                     ;X = Y + 1 if magic
.skip6512
        LDA $3B54,X             ;physic/magic evasion
        BEQ .skip5050           ;skip if null
        JSR $4B65               ;random: [0, A-1]
        INC
        STA $FF                 ;save evasion
.skip5050

    ;exit
        RTS

;-------------------------------------------------------------------------------
hit_adjust:
;
; adjust hit and evasion based on conditions
;-------------------------------------------------------------------------------

    ;init
        JSL flag_main           ;setup main flags
        LDA $EE                 ;flags

    ;miss if alive flag --------------------------------------------------------
        BIT #$04                ;miss if alive flag
        BNE save_miss

    ;miss if death immune flag -------------------------------------------------
        BIT #$02                ;miss if death protected flag
        BNE save_miss

    ;level x spell -------------------------------------------------------------

        ;check flag
            BIT #$40                ;lv x spell flag
            BEQ .skip8379           ;skip if not

        ;check hit rate
            TDC                     ;clean AB
            JSR $4B5A               ;random: [0, 255]
            LDX $11A8               ;hit rate
            BEQ save_miss
            CPX #$10                ;level x hit rate selector
            BCS .lvlx_high

        ;level x - low hit rate
.lvlx_low
            JSR $4792               ;C = C / X
            TXA                     ;get remainder
            BEQ save_hit            ;hit if remainder is zero
            BRA save_miss           ;miss otherwise

        ;level x - high hit rate
.lvlx_high
            CMP $11A8               ;hit rate
            BCC save_hit            ;hit if random is less than hit rate
            BRA save_miss           ;miss otherwise
.skip8379

    ;stamina -------------------------------------------------------------------

        ;check flag
            BIT #$10                ;stamina flag
            BEQ .skip6919           ;skip if not

        ;setup hit or miss
            LDA $3B40,Y             ;target stamina
            BEQ save_hit            ;setup hit if target stamina is null
            ASL                     ;stamina = stamina * 2
            BCC .skip_cap           ;skip if not overflow
            LDA #$FF                ;max stamina
.skip_cap            
            JSR $4B65               ;random; [0, A-1]
            INC
            STA $FF                 ;save as evasion

        ;exit
            RTS

.skip6919

    ;always hit ----------------------------------------------------------------
        BIT #$20                    ;always hit flag
        BNE save_hit

    ;select physic/magic conditions --------------------------------------------
        BIT #$01                    ;magic/physic
        BNE hit_adjust_phys         ;check physical conditions

;-------------------------------------------------------------------------------
hit_adjust_mag:
;
; adjust hit rate based on magic conditions
;-------------------------------------------------------------------------------

    ;init ----------------------------------------------------------------------
        JSL flag_mag                ;setup magic flags
        LDA $EE                     ;get it

    ;vanish --------------------------------------------------------------------
        BIT #$10                    ;vanish flag
        BNE save_hit

    ;reflect -------------------------------------------------------------------
        ;check flag
            BIT #$02                ;reflect flag
            BEQ .skip3393           ;skip if not

        ;setup reflect
            REP #$20                ;set C
            LDA $3018,Y             ;target id
            TSB $A6                 ;set as reflected
            SEP #$20                ;set A

        ;setup null animation
            LDA #$FF                ;null animation
            STA $EF                 ;save it

        ;setup miss
            BRA save_miss

.skip3393

    ;hit if petrify, muddle, sleep, stop, freeze -------------------------------
        ;check status
            PEA $A040                   ;petrify, muddle, sleep
            PEA $0210                   ;stop, freeze
            JSR $5864                   ;check target status
            BCC save_hit                ;hit if status are present
            LDA $EE                     ;reload flags

    ;beads 2 -------------------------------------------------------------------
        ;check flag
            BIT #$80                ;beads 2 flag
            BEQ .skip7126           ;skip if not

        ;check beads 2 chance
            JSR chance_mid          ;middle chance
            BCS .skip7126           ;skip if not

        ;setup miss
            BRA save_miss

.skip7126
        LDA $EE                     ;restore flags

    ;exit ----------------------------------------------------------------------
        RTS

;-------------------------------------------------------------------------------
save_func:
;
;save hit or miss functions
;-------------------------------------------------------------------------------

save_hit:
    LDA #$FF                        ;max hit rate
    BRA save_it

save_miss:
    TDC                             ;min hit rate
    
save_it:
    STA $FE                         ;save as hit rate
    RTS

;-------------------------------------------------------------------------------
hit_adjust_phys:
;
; - adjust physical hit based on conditions
;-------------------------------------------------------------------------------

    ;init ----------------------------------------------------------------------
        JSL flag_phys               ;setup physical flags
        LDA $EE                     ;get it

    ;golem ---------------------------------------------------------------------
        ;check flags
            BIT #$02                ;golem flag
            BEQ .skip6000           ;skip if not
            BIT #$08                ;single target
            BEQ .skip6000           ;skip if not

        ;decrement golem hits
            DEC $3A36               ;decrement golem hits to block by one

        ;setup animation
            LDA #$06
            STA $EF                 ;set golem animation

        ;setup golem byte
            STY $3A82               ;save character target in "golem blocked" byte

        ;setup hit
            BRA save_miss           ;set miss

.skip6000

    ;dog -----------------------------------------------------------------------
        ;check flags
            BIT #$40                ;dog
            BEQ .skip8999           ;skip if not
            BIT #$08                ;single target
            BEQ .skip8999           ;skip if not

        ;check chance
            JSR chance_low          ;low chance
            BCS .skip8999           ;skip if not

        ;setup dog block/animation
    	    LDA #$07
    	    STA $EF                 ;set dog animation
    	    STY $3A83               ;save character target in "Dog blocked" byte

        ;setup hit
    ;clean all attack status
            REP #$20                ;set C
            STZ $11AA               ;clean status
            STZ $11AC               ;clean status
            SEP #$20                ;set A
            JSR miss_dispel_2       ;null attack data
            BRA save_hit            ;save hit

.skip8999
        LDA $EE                     ;reload flags

    ;vanish --------------------------------------------------------------------

        ;check flags
            BIT #$10                ;vanish flag
            BNE save_miss

    ;image ---------------------------------------------------------------------

        ;check flags
            BIT #$20                ;image flag
            BEQ .skip4278           ;skip if not

        ;check if attack setups image
            LDA $11AB               ;status flags
            BIT #$04                ;image status
            BNE .skip_dispel        ;skip if attack setups image status 

        ;check chance for dispel
            JSR chance_mid          ;middle chance
            BCS .skip_dispel        ;skip if outside chance

        ;remove image status
            LDA $3DFD,Y             ;status to dispel
            ORA #$04                ;add image status to dispel
            STA $3DFD,Y             ;update status to dispel

        .skip_dispel
        ;setup miss
            BRA save_miss

.skip4278

    ;petrify, muddle, sleep, stop, freeze --------------------------------------

        ;check status and setup hit
            PEA $A040               ;petrify, muddle, sleep
            PEA $0210               ;stop, freeze
            JSR $5864               ;check target status
            BCC save_hit            ;setup hit if status set
            LDA $EE                 ;reload flags

    ;blind ---------------------------------------------------------------------

        ;check flags
            BIT #$01                ;blind flag
            BEQ .skip5720           ;skip if not

        ;check rate
            JSR chance_hig          ;high chance
            BCC save_hit

.skip5720
            LDA $EE                 ;reload flags

    ;beads 1 -------------------------------------------------------------------

        ;check flags
            BIT #$04                ;beads 1
            BEQ .skip4216           ;skip if not

        ;check rate
            JSR chance_mid          ;middle chance
            BCS .skip4216

        ;setup animation
            LDA #$05
            STA $EF                 ;setup beads animation

        ;setup miss
            JMP save_miss

.skip4216
            LDA $EE                 ;restore flags

    ;exit ----------------------------------------------------------------------

        RTS

;-------------------------------------------------------------------------------
hit_define:
;
;setup hit or miss based on hit and evasion rate
;-------------------------------------------------------------------------------

    ;check min/max values
        LDA $FE                     ;hit rate
        BEQ .save_miss              ;setup miss if hit rate is zero
        CMP #$FF
        BEQ .save_hit               ;setup hit if hit rate is max

    ;calculate hit or miss
        CMP $FF                     ;compare with evasion
        BCS .save_hit               ;setup hit if hit rate >= evasion
        BRA .save_miss              ;miss otherwise

    .save_hit:
        JMP save_hit

    .save_miss:
        JMP save_miss

;-------------------------------------------------------------------------------
miss_animation:
;
;setup miss animations and special animations
;-------------------------------------------------------------------------------

    ;init
        JSL flag_main               ;setup flags

    ;check if ally
        CPY #$08                    ;max ally index identifier
        BCS .exit                   ;exit if monster index

    ;check if physical 
        LDA $EE                     ;flags
        BIT #$01                    ;magic or physic attack
        BEQ .exit                   ;exit if magic attack

    ;check if animation was defined
        LDA $EF                     ;animation
        BNE .setup                  ;setup animation if defined

    ;check if attack hits
        LDA $FE                     ;hit rate
        BMI .exit                   ;exit if attack hits

    ;check if stamina based attack
        LDA $EE                     ;flags
        BIT #$10                    ;stamina flag
        BNE .exit                   ;exit if stamina based attack

    ;check status that supress animations:
        PEA $A45A                   ;zombie, magitek, vanish, petrify, image, muddle, sleep
        PEA $0210                   ;stop, freeze
        JSR $5864                   ;check target status
        BCC .exit                   ;don't set animation for those status
    
    ;find animations from equipment
        LDA $3CE4,Y                 ;load animation of available equipment
        BEQ .exit                   ;exit if no equipment animation

    ;get animation value
        JSR $522A                   ;choose a random bit
        JSR $51F0                   ;convert bit set to [0,7] value. Store at X
        TXA
        INC

    ;setup animation
    .setup
        BMI .exit                   ;don't store animation if #$FF
        PHA                         ;save value for animation
        TYA                         ;get fighter index
        LSR                         ;adjust from [0,2,4,6] index to [0,1,2,3] index
        TAX                         ;save index
        PLA                         ;retrieve animation
        STA $00AA,X                 ;setup evasion animation for fighter

    ;exit
    .exit
        RTS

;-------------------------------------------------------------------------------
miss_dispel:
;
; dispel status and clean attack data when miss
;-------------------------------------------------------------------------------

    ;check if attack hits
        LDA $FE                 ;hit rate
        BMI md_exit             ;exit if attack hits

miss_dispel_2:

    ;setup no critical and ignore true knight
        LDA #$02                ;no critical and ignore true knight
        TSB $B2                 ;save it

    ;turn off random weapon spellcast
        STZ $3A89               ;turn off random weapon spellcast

    ;check if missable weapon spellcast
        LDA $341C
        BEQ md_exit            ;exit if current strike is missable weapon spellcast

    ;check special animation
        LDA $EF                 ;animation
        BMI md_exit             ;exit for null animation ($EF = #$FF)

    ;setup miss animation
        REP #$20                ;set C
        LDA $3018,Y             ;target bit identifier
        TSB $3A5A               ;setup target as missed
        SEP #$20                ;set A
    
md_exit:

    ;exit
        RTS

;-------------------------------------------------------------------------------
hit_carry:
;
; setup carry for callers
;-------------------------------------------------------------------------------

    ;check if hit or miss
        LDA $FE                 ;hit rate
        BMI .setup_hit          ;set hit

    .setup_miss
        SEC
        RTS

    .setup_hit
        CLC
        RTS

warnpc  $C223EC

;===============================================================================
;
; - flags setup
;
;===============================================================================
org     $C3FD00                 ;(*) change for the start of the free space

;-------------------------------------------------------------------------------
flag_main:
;
; setup flags for main hit conditions:
;
; 01 physic
; 02 miss if death immune
; 04 miss if alive
; 08 ----
; 10 stamina
; 20 can't dodge
; 40 lv X spell
; 80 ----
;-------------------------------------------------------------------------------

    ;reset flags
        STZ $EE                 ;reset flags

    ;setup physic
        LDA $11A2               ;attack flags
        AND #$01                ;physic attack
        TSB $EE

    ;setup death immune
        LDA $11A2               ;attack flags
        BIT #$02                ;spell miss if death protected flag
        BEQ .skip6215           ;skip if not
        LDA $3AA1,Y             ;?
        BIT #$04                ;death immune
        BEQ .skip6215           ;skip if not
        LDA #$02                ;death immune flag
        TSB $EE
.skip6215

    ;setup alive
        LDA $11A2               ;attack flags
        BIT #$04                ;attack: spell miss if not dead targets
        BEQ .skip8123           ;skip if not
        LDA $3EE4,Y             ;target status
        AND #$80                ;death status
        BNE .skip8123
        LDA #$04                ;alive flag
        TSB $EE
.skip8123

    ;setup lv x spell, stamina, can't dodge
        LDA $11A4               ;attack flags
        AND #$70                ;lv x spell, stamina, can't dodge
        TSB $EE

    ;exit
        RTL

;-------------------------------------------------------------------------------
flag_phys:
;
; setup flags for physical hit conditions:
;
; 01 blind
; 02 golem
; 04 beads 1
; 08 single
; 10 vanish
; 20 image
; 40 dog
; 80 ----
;-------------------------------------------------------------------------------

    ;reset flags
        STZ $EE                 ;reset flags

    ;setup blind, vanish
        LDA $3EE4,Y             ;target status
        AND #$11                ;blind, vanish
        TSB $EE

    ;setup golem
        LDA $3A36               ;golem hits
        BEQ .skip8512           ;skip if no golem effect
        LDA #$02                ;golem flag
        TSB $EE
    .skip8512

    ;setup beads 1
        LDA $3C58,Y             ;flags
        AND #$04                ;beads 1 flag
        TSB $EE    

    ;setup single
        LDA $3EC9               ;number of targets
        CMP #$01
        BNE .skip0099           ;skip if more than one target
        LDA $11A9               ;special effect. They can change hit rate or targets
        BNE .skip0099           ;skip if number of targets can change
        LDA #$08                ;single flag
        TSB $EE
    .skip0099

    ;setup image
        LDA $3EE5,Y             ;target status
        AND #$04                ;image
        BEQ .skip3714           ;skip if not
        LDA #$20                ;image flag
        TSB $EE
    .skip3714

    ;setup dog
        LDA $3EF9,Y             ;target status
        AND #$40                ;dog
        TSB $EE

    ;exit
        RTL

;-------------------------------------------------------------------------------
flag_mag:
;
; setup flags for magical hit conditions:
;
; 01 ----
; 02 reflect
; 04 ----
; 08 ----
; 10 vanish
; 20 ----
; 40 ----
; 80 beads 2
;------------------------------------------------------------------------------- 

    ;reset flags
        STZ $EE                 ;reset flags

    ;setup reflect
        LDA $11A3               ;attack fkags
        AND #$02                ;ignore reflect
        BNE .skip6149           ;skip if attack ignores reflect
        LDA $3EF8,Y             ;target status
        AND #$80                ;reflect
        BEQ .skip6149           ;skip if not reflect
        LDA #$02                ;reflect flag
        TSB $EE
    .skip6149

    ;setup vanish
        LDA $3EE4,Y             ;target status
        AND #$10                ;vanish
        TSB $EE

    ;setup beads 2
        LDA $3C58,Y             ;flags
        AND #$80                ;beads 2 flag
        TSB $EE

    ;exit
        RTL

;-------------------------------------------------------------------------------
hit_dispel:
;
; clean status when hit
;-------------------------------------------------------------------------------

    ;check ignore reaction flag
        LDA $B3                 ;flags
        BPL .exit               ;exit if ignore reaction

    ;check healing 
        LDA $11A4               ;attack bits
        BIT #$01                ;heal flag
        BNE .exit               ;exit if it heals

    ;check attack power
        LDA $11A6               ;attack power
        BEQ .exit               ;skip if attack power is zero

    ;check golem/dog
        LDA $3A82
        AND $3A83
        BPL .exit               ;skip if golem or dog

    ;remove control
        PHX                     ;backup X
        TYX                     ;X = Y
        ASL $32B8,X
        SEC
        ROR $32B8,X             ;flag to clear: who you controls
        ASL $32B9,X 
        SEC 
        ROR $32B9,X             ;flag to clear: who controls you
        PLX                     ;restore X

    ;remove status
        REP #$20                ;mode 2AB
        LDA $11AA               ;attack status
        EOR #$A010              ;vanish, muddle, sleep
        AND #$A010              
        AND $3EE4,Y             ;status to setup
        ORA $3DFC,Y             ;status to clean
        STA $3DFC,Y             ;update it

    ;exit
    .exit
        SEP #$20                ;set 1AB
        RTL

warnpc  $C3FDD4                 ;(*) change for the start of the free space

